from __future__ import annotations

import dataclasses
import functools
import inspect
import itertools
import textwrap
from collections.abc import Iterable
from functools import partial
from typing import TYPE_CHECKING, Callable

from pydiverse.transform.core import dtypes
from pydiverse.transform.errors import ExpressionTypeError

if TYPE_CHECKING:
    from pydiverse.transform.ops import Operator, OperatorExtension


class OperatorImpl:
    """
    Internal type to store the implementation of an operator and all associated
    metadata.

    If the function (`impl`) provided as the underlying parameter has keyword
    only arguments that start with an underscore, they get added to the
    `self.internal_kwargs` list.
    """

    id_ = itertools.count()

    def __init__(
        self,
        operator: Operator,
        impl: Callable,
        signature: OperatorSignature,
    ):
        self.operator = operator
        self.impl = impl
        self.signature = signature
        self.variants: dict[str, Callable] = {}

        self.__id = next(OperatorImpl.id_)

        # Inspect impl signature to get internal kwargs
        self.internal_kwargs = self._compute_internal_kwargs(impl)
        self.variant_internal_kwargs = {}

        # Calculate Ordering Key
        # - Find match with the least number templates in the signature
        # - From those take the one with the least number of different templates
        # - From those take the one where the first template appears latest
        # - From those take the one where the const modifiers match better
        # - From those take the one that isn't a vararg or has the most arguments.
        # - From those take the one that was defined first

        num_templates = 0
        templates_set = set()
        template_indices = []
        const_modifiers = []
        for i, dtype in enumerate(signature.args):
            if isinstance(dtype, dtypes.Template):
                num_templates += 1
                templates_set.add(dtype.name)
                template_indices.append(-i)
            if dtype.const:
                const_modifiers.append(i)
        num_different_templates = len(templates_set)
        is_vararg = int(self.signature.is_vararg)

        self._precedence = (
            num_templates,
            num_different_templates,
            tuple(template_indices),
            tuple(const_modifiers),
            is_vararg,
            -len(signature.args),
            self.__id,
        )

    def add_variant(self, name: str, impl: Callable):
        if name in self.variants:
            raise ValueError(
                f"Already added a variant with name '{name}'"
                f" to operator {self.operator}."
            )
        self.variants[name] = impl
        self.variant_internal_kwargs[name] = self._compute_internal_kwargs(impl)

    @staticmethod
    def _compute_internal_kwargs(impl: Callable):
        internal_kwargs = []
        try:
            impl_signature = inspect.signature(impl)
            for name, param in impl_signature.parameters.items():
                if param.kind == inspect.Parameter.KEYWORD_ONLY and name.startswith(
                    "_"
                ):
                    internal_kwargs.append(name)
        except (TypeError, ValueError):
            pass

        return internal_kwargs


@dataclasses.dataclass
class TypedOperatorImpl:
    """
    Operator Implementation with a non-templated return type.
    Unlike `OperatorImpl`, this class is intended to be the return type of
    the OperatorRegistry.
    """

    operator: Operator
    impl: OperatorImpl
    rtype: dtypes.DType

    @classmethod
    def from_operator_impl(cls, impl: OperatorImpl, rtype: dtypes.DType):
        return cls(
            operator=impl.operator,
            impl=impl,
            rtype=rtype,
        )

    def __call__(self, *args, **kwargs):
        return self.impl.impl(*args, **self.__clean_kwargs(kwargs))

    def has_variant(self, name: str) -> bool:
        return name in self.impl.variants

    def get_variant(self, name: str) -> Callable | None:
        variant = self.impl.variants.get(name)
        if variant is None:
            return None

        @functools.wraps(variant)
        def variant_wrapper(*args, **kwargs):
            return variant(*args, **self.__clean_kwargs(kwargs, variant=name))

        return variant_wrapper

    def __clean_kwargs(self, kwargs, variant=None):
        internal_kwargs = self.impl.internal_kwargs
        if variant:
            internal_kwargs = self.impl.variant_internal_kwargs[variant]

        return {
            k: v
            for k, v in kwargs.items()
            if not k.startswith("_") or k in internal_kwargs
        }


class OperatorRegistry:
    # It only makes sense to define some dunder methods.
    # These are the ones which can be registered.
    SUPPORTED_DUNDER = {
        "__add__",
        "__radd__",
        "__sub__",
        "__rsub__",
        "__mul__",
        "__rmul__",
        "__truediv__",
        "__rtruediv__",
        "__floordiv__",
        "__rfloordiv__",
        "__pow__",
        "__rpow__",
        "__mod__",
        "__rmod__",
        "__round__",
        "__pos__",
        "__neg__",
        "__abs__",
        "__and__",
        "__rand__",
        "__or__",
        "__ror__",
        "__xor__",
        "__rxor__",
        "__invert__",
        "__lt__",
        "__le__",
        "__eq__",
        "__ne__",
        "__gt__",
        "__ge__",
    }

    # Set containing all operator names that have been defined across all registries.
    # Used for __dir__ method of SymbolicExpression
    ALL_REGISTERED_OPS: set[str] = set()

    def __init__(self, name, super_registry=None):
        self.name = name
        self.super_registry = super_registry
        self.registered_ops = set()  # type: set[Operator]
        self.implementations = dict()  # type: dict[str, OperatorImplementationStore]
        self.check_super = dict()  # type: dict[str, bool]

    def register_op(self, operator: Operator, check_super=True):
        """
        :param operator: The operator to register.
        :param check_super: Bool indicating if the super register should be
            checked if no implementation for this operator can be found.
        """

        name = operator.name
        if name.startswith("__") and name.endswith("__"):
            if name not in OperatorRegistry.SUPPORTED_DUNDER:
                raise ValueError(f"Dunder method {name} is not supported.")

        if operator in self.registered_ops:
            raise ValueError(
                f"Operator {operator} ({name}) already registered in this operator"
                f" registry '{self.name}'"
            )
        if name in self.implementations:
            raise ValueError(
                f"Another operator with the name '{name}' has already been registered"
                " in this registry."
            )

        self.implementations[name] = OperatorImplementationStore(operator)
        self.check_super[name] = check_super

        self.registered_ops.add(operator)
        self.ALL_REGISTERED_OPS.add(name)

    def get_operator(self, name: str) -> Operator | None:
        if impl_store := self.implementations.get(name, None):
            return impl_store.operator

        # If operation hasn't been defined in this registry, go to the parent
        # registry and check if it has been defined there.
        if self.super_registry is None or not self.check_super.get(name, True):
            raise ValueError(f"No implementation for operator '{name}' found")
        return self.super_registry.get_operator(name)

    def add_implementation(
        self,
        operator: Operator,
        impl: Callable,
        signature: str,
        variant: str | None = None,
    ):
        if operator not in self.registered_ops:
            raise ValueError(
                f"Operator {operator} ({operator.name}) hasn't been registered in this"
                f" operator registry '{self.name}'"
            )

        signature = OperatorSignature.parse(signature)
        operator.validate_signature(signature)

        implementation_store = self.implementations[operator.name]
        op_impl = OperatorImpl(operator, impl, signature)

        if variant:
            implementation_store.add_variant(variant, op_impl)
        else:
            implementation_store.add_implementation(op_impl)

    def get_implementation(self, name, args_signature) -> TypedOperatorImpl:
        if name not in self.ALL_REGISTERED_OPS:
            raise ValueError(f"No operator named '{name}'.")

        for dtype in args_signature:
            if not isinstance(dtype, dtypes.DType):
                raise TypeError(
                    "Expected elements of `args_signature` to be of type DType,"
                    f" but found element of type {type(dtype).__name__} instead."
                )

        if store := self.implementations.get(name):
            if impl := store.find_best_match(args_signature):
                return impl

        # If operation hasn't been defined in this registry, go to the parent
        # registry and check if it has been defined there.
        if self.super_registry is None or not self.check_super.get(name, True):
            raise ValueError(
                f"No implementation for operator '{name}' found that matches signature"
                f" '{args_signature}'."
            )
        return self.super_registry.get_implementation(name, args_signature)


class OperatorSignature:
    """
    Specification:

        signature ::= arguments "->" rtype
        arguments ::= (modifiers dtype ",")* terminal_arg
        terminal_arg ::= modifiers (dtype | vararg)
        vararg ::= dtype "..."
        rtype ::= dtype
        dtype ::= template | "int" | "float" | "str" | "bool" | and others...
        modifiers ::= "const"?
        template ::= single uppercase character

    Examples:

        Function that takes two integers and returns an integer:
            int, int -> int

        Templated argument (templates consist of single uppercase characters):
            T, T -> T
            T, U -> bool

        Variable number of arguments:
            int... -> int

    """

    def __init__(self, args: list[dtypes.DType], rtype: dtypes.DType):
        """
        :param args: Tuple of argument types.
        :param rtype: The return type.
        """
        self.args = args
        self.rtype = rtype

    @classmethod
    def parse(cls, signature: str) -> OperatorSignature:
        def parse_cstypes(cst: str):
            # cstypes = comma seperated types
            types = cst.split(",")
            types = [t.strip() for t in types]
            types = [dtypes.dtype_from_string(t) for t in types if t]
            return types

        if "->" not in signature:
            raise ValueError("Invalid signature: arrow (->) missing.")

        arg_sig, r_sig = signature.split("->")
        args = parse_cstypes(arg_sig)
        rtype = parse_cstypes(r_sig)

        # Validate Signature
        if len(rtype) != 1:
            raise ValueError(
                f"Invalid operator signature '{signature}'. Expected exactly one return"
                " type."
            )

        rtype = rtype[0]

        # Validate Template
        # Output template must also occur in signature
        if isinstance(rtype, dtypes.Template):
            if rtype.name not in [arg.name for arg in args]:
                raise ValueError(
                    f"Template return type '{rtype}' must also occur in the"
                    " argument signature."
                )

        # Validate vararg
        # Vararg can only occur for the very last element
        for arg in args[:-1]:
            if arg.vararg:
                raise ValueError("Only last argument can be a vararg.")

        if rtype.vararg:
            raise ValueError("Return value can't be a vararg.")

        return OperatorSignature(
            args=args,
            rtype=rtype,
        )

    def __repr__(self):
        args_str = ", ".join(str(arg) for arg in self.args)
        return args_str + " -> " + str(self.rtype)

    def __hash__(self):
        return hash((self.args, self.rtype))

    def __eq__(self, other):
        if not isinstance(other, OperatorSignature):
            return False
        return self.args == other.args and self.rtype == other.rtype

    @property
    def is_vararg(self) -> bool:
        if len(self.args) == 0:
            return False
        return self.args[-1].vararg


class OperatorImplementationStore:
    """
    Stores all implementations for a specific operation in a trie according to
    their signature. This enables us to easily find the best matching
    operator implementation for a given set of input argument types.
    """

    @dataclasses.dataclass
    class TrieNode:
        __slots__ = ("value", "operator", "children")
        value: dtypes.DType
        operator: OperatorImpl | None
        children: list[OperatorImplementationStore.TrieNode]

        def __repr__(self):
            self_text = f"({self.value} - {self.operator})"
            if self.children:
                children_text = "\n".join(repr(c) for c in self.children)
                children_text = textwrap.indent(children_text, "  ")
                return self_text + "\n" + children_text
            return self_text

    def __init__(self, operator: Operator):
        self.operator = operator
        self.root = self.TrieNode("ROOT", None, [])  # type: ignore

    def add_implementation(self, operator: OperatorImpl):
        node = self.get_node(operator.signature, create_missing=True)
        if node.operator is not None:
            raise ValueError(
                f"Implementation for signature {operator.signature} already defined."
            )

        node.operator = operator

    def add_variant(self, name: str, operator: OperatorImpl):
        node = self.get_node(operator.signature, create_missing=False)
        if node is None or node.operator is None:
            raise ValueError(
                f"No implementation for signature {operator.signature} found."
                " Make sure there is an exact match to add a variant."
            )

        assert node.operator.signature.rtype.same_kind(operator.signature.rtype)
        node.operator.add_variant(name, operator.impl)

    def get_node(self, signature: OperatorSignature, create_missing: bool = True):
        node = self.root
        for dtype in signature.args:
            for child in node.children:
                if child.value == dtype:
                    node = child
                    break
            else:
                if create_missing:
                    new_node = self.TrieNode(dtype, None, [])
                    node.children.append(new_node)
                    node = new_node
                else:
                    return None

        return node

    def find_best_match(
        self, signature: tuple[dtypes.DType]
    ) -> TypedOperatorImpl | None:
        matches = list(self._find_matches(signature))

        if not matches:
            return None

        # Find best matching template.
        best_match = None
        best_score = ((0x7FFFFFFF,), (0x7FFFFFFF,))

        for match, templates, type_promotion_indices in matches:
            score = (
                # Prefer operators that didn't need any type promotion
                tuple(-i for i in type_promotion_indices),
                # And then match according to signature
                match.operator._precedence,
            )
            if score < best_score:
                best_match = (match, templates)
                best_score = score

        rtype = best_match[0].operator.signature.rtype
        if isinstance(rtype, dtypes.Template):
            # If rtype is a template -> Translate
            rtype = best_match[1][rtype.name]

        return TypedOperatorImpl.from_operator_impl(best_match[0].operator, rtype)

    def _find_matches(
        self, signature: tuple[dtypes.DType]
    ) -> Iterable[TrieNode, dict[str, dtypes.DType, tuple[int, ...]]]:
        """Yield all operators that match the input signature"""

        # Case 0 arguments:
        if len(signature) == 0:
            yield self.root, dict(), tuple()
            return

        # Case 1+ args:
        def does_match(
            dtype: dtypes.DType,
            node: OperatorImplementationStore.TrieNode,
        ) -> bool:
            if isinstance(node.value, dtypes.Template):
                return node.value.modifiers_compatible(dtype)
            return dtype.can_promote_to(node.value)

        stack: list[
            tuple[OperatorImplementationStore.TrieNode, int, dict, tuple[int, ...]]
        ] = [(child, 0, dict(), tuple()) for child in self.root.children]

        while stack:
            node, s_i, templates, type_promotion_indices = stack.pop()
            dtype = signature[s_i]

            if not does_match(dtype, node):
                continue

            if isinstance(node.value, dtypes.Template):
                templates = templates.copy()
                if node.value.name not in templates:
                    templates[node.value.name] = [dtype.without_modifiers()]
                else:
                    templates[node.value.name] = templates[node.value.name] + [
                        dtype.without_modifiers()
                    ]
            elif not node.value.same_kind(dtype):
                # Needs type promotion
                # This only works when types can be promoted once
                # -> (uint > int) wouldn't be preferred over (uint > int > float)
                type_promotion_indices = (*type_promotion_indices, s_i)

            if s_i + 1 == len(signature):
                if node.operator is not None:
                    # Find compatible type for templates
                    try:
                        templates = {
                            name: dtypes.promote_dtypes(types_)
                            for name, types_ in templates.items()
                        }
                        yield node, templates, type_promotion_indices
                    except ExpressionTypeError:
                        print(f"Can't promote: {templates}")
                        pass

                continue

            children = iter(node.children)
            if node.value.vararg:
                children = itertools.chain(children, iter((node,)))

            for child in children:
                stack.append((child, s_i + 1, templates, type_promotion_indices))


class OperatorRegistrationContextManager:
    def __init__(
        self,
        registry: OperatorRegistry,
        operator: Operator,
        *,
        check_super=True,
    ):
        self.registry = registry
        self.operator = operator

        self.registry.register_op(operator, check_super=check_super)

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        pass

    def __call__(self, signature: str, *, variant: str = None):
        if not isinstance(signature, str):
            raise TypeError("Signature must be of type str.")

        def decorator(func):
            self.registry.add_implementation(
                self.operator,
                func,
                signature,
                variant,
            )
            return func

        return decorator

    def auto(self, func: Callable = None, *, variant: str = None):
        if func is None:
            return partial(self.auto, variant=variant)

        if not self.operator.signatures:
            raise ValueError(f"Operator {self.operator} has not default signatures.")

        for signature in self.operator.signatures:
            self.registry.add_implementation(
                self.operator,
                func,
                signature,
                variant,
            )

        return func

    def extension(self, extension: type[OperatorExtension], variant: str = None):
        if extension.operator != type(self.operator):  # noqa: E721
            raise ValueError(
                f"Operator extension for '{extension.operator.__name__}' can't "
                f"be applied to operator of type '{type(self.operator).__name__}'."
            )

        def decorator(func):
            for sig in extension.signatures:
                self.registry.add_implementation(self.operator, func, sig, variant)
            return func

        return decorator



from __future__ import annotations

import functools
import itertools
import operator as py_operator
import uuid
import warnings
from typing import Any, Callable

import numpy as np
import pandas as pd
import pandas.core.dtypes.cast
import pandas.core.dtypes.dtypes

from pydiverse.transform import ops
from pydiverse.transform.core import dtypes
from pydiverse.transform.core.expressions import (
    Column,
    FunctionCall,
    LiteralColumn,
    SymbolicExpression,
)
from pydiverse.transform.core.expressions.translator import Translator, TypedValue
from pydiverse.transform.core.util import OrderingDescriptor, translate_ordering
from pydiverse.transform.eager.eager_table import EagerTableImpl, uuid_to_str
from pydiverse.transform.errors import (
    AlignmentError,
    ExpressionError,
    FunctionTypeError,
)
from pydiverse.transform.ops import OPType

__all__ = [
    "PandasTableImpl",
]


class PandasTableImpl(EagerTableImpl):
    """Pandas backend

    Attributes:
        df: The current, eager dataframe. It's columns get renamed to prevent
            name collisions, and it also contains all columns that have been
            computed at some point. This allows for a more lazy style API.
    """

    def __init__(self, name: str, df: pd.DataFrame):
        self.df = fast_pd_convert_dtypes(df)
        self.join_translator = self.JoinTranslator()

        columns = {
            name: Column(name=name, table=self, dtype=self._dtype_from_pd(dtype))
            for name, dtype in self.df.dtypes.items()
        }

        # Rename columns
        self.df_name_mapping = {
            col.uuid: f"{name}_{col.name}_" + uuid_to_str(col.uuid)
            for col in columns.values()
        }  # type: dict[uuid.UUID: str]

        self.df = self.df.rename(
            columns={
                name: self.df_name_mapping[col.uuid] for name, col in columns.items()
            }
        )

        super().__init__(name=name, columns=columns)

    def _dtype_from_pd(self, type_: Any) -> dtypes.DType:
        if pd.api.types.is_integer_dtype(type_):
            return dtypes.Int()
        if pd.api.types.is_float_dtype(type_):
            return dtypes.Float()
        if pd.api.types.is_bool_dtype(type_):
            return dtypes.Bool()
        if pd.api.types.is_string_dtype(type_):
            return dtypes.String()
        if pd.api.types.is_datetime64_any_dtype(type_):
            return dtypes.DateTime()

        raise NotImplementedError(f"Unsupported type: {type_}")

    def _dtype_to_pd(self, dtype: dtypes.DType) -> Any:
        if dtypes.Int().same_kind(dtype):
            return pd.Int64Dtype()
        if dtypes.Float().same_kind(dtype):
            return pd.Float64Dtype()
        if dtypes.Bool().same_kind(dtype):
            return pd.BooleanDtype()
        if dtypes.String().same_kind(dtype):
            return pd.StringDtype()
        if dtypes.DateTime().same_kind(dtype):
            return "datetime64[ns]"

        raise NotImplementedError(f"Invalid type: {dtype}")

    def is_aligned_with(self, col) -> bool:
        len_self = len(self.df.index)

        if isinstance(col, Column):
            if not isinstance(col.table, type(self)):
                return False
            return len(col.table.df.index) == len_self

        if isinstance(col, LiteralColumn):
            if not issubclass(col.backend, type(self)):
                return False
            if isinstance(col.typed_value.value, pd.Series):
                return len(col.typed_value.value) == len_self
            return True  # A constant value (eg int, float, str...)

        raise ValueError

    #### Verb Operations ####

    def alias(self, name=None):
        # Creating a new table object also acts like a garbage collecting mechanism.
        new_name = name or self.name
        return self.__class__(new_name, self.collect())

    def collect(self) -> pd.DataFrame:
        # SELECT -> Apply mask
        selected_cols_name_map = {
            self.df_name_mapping[uuid]: name for name, uuid in self.selected_cols()
        }
        masked_df = self.df[[*selected_cols_name_map.keys()]]

        # rename columns from internal naming scheme to external names
        result = masked_df.rename(columns=selected_cols_name_map)

        # Add metadata
        result.attrs["name"] = self.name
        return result

    def mutate(self, **kwargs):
        uuid_kwargs = {self.named_cols.fwd[k]: (k, v) for k, v in kwargs.items()}
        self.df_name_mapping.update(
            {
                uuid: f"{self.name}_mutate_{name}_" + uuid_to_str(uuid)
                for uuid, (name, _) in uuid_kwargs.items()
            }
        )

        # Update Dataframe
        grouper = self.df_grouper()
        cols = {
            self.df_name_mapping[uuid]: self.cols[uuid].compiled(
                self.df, grouper=grouper
            )
            for uuid in uuid_kwargs.keys()
        }

        self.df = self.df.copy(deep=False)
        for col_name, col in cols.items():
            self.df[col_name] = col
        self.df = fast_pd_convert_dtypes(self.df)

    def join(
        self,
        right: PandasTableImpl,
        on: SymbolicExpression,
        how: str,
        *,
        validate=None,
    ):
        """
        :param on: Symbolic expression consisting of anded column equality checks.
        """

        # Parse ON condition
        on_cols = []
        for col1, col2 in self.join_translator.translate(on):
            if col1.uuid in self.cols and col2.uuid in right.cols:
                on_cols.append((col1, col2))
            elif col2.uuid in self.cols and col1.uuid in right.cols:
                on_cols.append((col2, col1))
            else:
                raise Exception

        if validate is not None:
            pd_validate = {"1:?": "1:1"}[validate]
        else:
            pd_validate = None

        # Perform Join
        left_on, right_on = zip(*on_cols)
        left_on = [self.df_name_mapping[col.uuid] for col in left_on]
        right_on = [right.df_name_mapping[col.uuid] for col in right_on]

        self.df_name_mapping.update(right.df_name_mapping)

        # TODO: Implement more merge operations that preserve the index / ordering.
        if how == "left" and validate == "1:?":
            # Perform an order and index preserving join
            original_index = self.df.index
            tmp_df = self.df.reset_index(drop=True)
            merged_df = tmp_df.merge(
                right.df,
                how=how,
                left_on=left_on,
                right_on=right_on,
                validate=pd_validate,
            ).loc[tmp_df.index]
            merged_df.index = original_index
            self.df = merged_df
        else:
            self.df = self.df.merge(
                right.df,
                how=how,
                left_on=left_on,
                right_on=right_on,
                validate=pd_validate,
            )

    def filter(self, *args: SymbolicExpression):
        if not args:
            return

        compiled, dtype = self.compiler.translate(
            functools.reduce(py_operator.and_, args)
        )
        assert isinstance(dtype, dtypes.Bool)

        condition = compiled(self.df)
        self.df = self.df.loc[condition]

    def arrange(self, ordering):
        self.df = self.sort_df(self.df, ordering)

    def summarise(self, **kwargs):
        uuid_kwargs = {self.named_cols.fwd[k]: (k, v) for k, v in kwargs.items()}
        self.df_name_mapping.update(
            {
                uuid: f"{self.name}_summarise_{name}_" + uuid_to_str(uuid)
                for uuid, (name, _) in uuid_kwargs.items()
            }
        )

        # Update Dataframe
        grouper = self.grouped_df().grouper if self.grouped_by else None
        aggr_cols = {
            self.df_name_mapping[uuid]: self.cols[uuid].compiled(
                self.df, grouper=grouper
            )
            for uuid in uuid_kwargs.keys()
        }

        # Grouped Dataframe requires different operations compared to ungrouped df.
        if self.grouped_by:
            columns = {
                k: fast_pd_convert_dtypes(v.rename(k)) for k, v in aggr_cols.items()
            }
            self.df = pd.concat(columns, axis="columns", copy=False).reset_index()
        else:
            self.df = fast_pd_convert_dtypes(pd.DataFrame(aggr_cols, index=[0]))

    def slice_head(self, n: int, offset: int):
        self.df = self.df[offset : n + offset]

    #### EXPRESSIONS ####

    class ExpressionCompiler(
        EagerTableImpl.ExpressionCompiler[
            "PandasTableImpl", TypedValue[Callable[[pd.DataFrame], pd.Series]]
        ]
    ):
        def _translate_col(self, expr, **kwargs):
            df_col_name = self.backend.df_name_mapping[expr.uuid]

            def df_col(df, **kw):
                return df[df_col_name]

            return TypedValue(df_col, expr.dtype)

        def _translate_literal_col(self, expr, **kwargs):
            if not self.backend.is_aligned_with(expr):
                raise AlignmentError(
                    "Literal column isn't aligned with this table. "
                    f"Literal Column: {expr}"
                )

            def value(df, **kw):
                literal_value = expr.typed_value.value
                if isinstance(literal_value, pd.Series):
                    literal_len = len(literal_value.index)
                    if len(df.index) != literal_len:
                        raise AlignmentError(
                            "Literal column isn't aligned with this table. Make sure"
                            f" that they have the same length ({len(df.index)} vs"
                            f" {literal_len}). This might be the case if you are using"
                            " literal columns inside a window function.\nLiteral"
                            f" Column: {expr}"
                        )

                    series = literal_value.copy()
                    series.index = df.index
                    return series
                return literal_value

            return TypedValue(value, expr.typed_value.dtype, expr.typed_value.ftype)

        def _translate_function(
            self, expr, implementation, op_args, context_kwargs, *, verb=None, **kwargs
        ):
            internal_kwargs = {}

            def value(df, *, grouper=None, **kw):
                result_dtype = self.backend._dtype_to_pd(implementation.rtype)

                args = [arg.value(df, grouper=grouper, **kw) for arg in op_args]
                kwargs = {
                    "_tbl": self.backend,
                    "_df": df,
                    "_result_dtype": result_dtype,
                    **internal_kwargs,
                }

                # Element wise operator
                # -> Perform on ungrouped columns
                if implementation.operator.ftype == OPType.EWISE:
                    return implementation(*args, **kwargs)

                # Get arguments; If there are no arguments, use dataframe index
                # instead for grouping.
                if len(args) > 0:
                    series, *args = args
                else:
                    series = df.index.to_series()

                def bound_impl(x):
                    """
                    Helper that binds the *args and **kwargs to the implementation.
                    Takes one argument - the rest is already supplied.
                    """
                    return implementation(x, *args, **kwargs)

                # Grouped
                if grouper is not None:
                    series = series.groupby(grouper)

                    if verb == "summarise":
                        if variant := implementation.get_variant("aggregate"):
                            return variant(series, *args, **kwargs)
                        return series.aggregate(bound_impl)
                    if variant := implementation.get_variant("transform"):
                        return variant(series, *args, **kwargs)
                    return series.transform(bound_impl)

                # Ungrouped
                scalar = bound_impl(series)
                if verb == "summarise":
                    return scalar

                return pd.Series(scalar, index=series.index, dtype=result_dtype)

            operator = implementation.operator

            if operator.ftype == OPType.WINDOW:
                if verb != "mutate":
                    raise FunctionTypeError(
                        "Window function are only allowed inside a mutate."
                    )

                if arrange := context_kwargs.get("arrange"):
                    ordering = translate_ordering(self.backend, arrange)
                    internal_kwargs["_ordering"] = ordering

                value = self.arranged_window(
                    value, operator, context_kwargs, verb=verb, **kwargs
                )

            override_ftype = (
                OPType.WINDOW
                if operator.ftype == OPType.AGGREGATE and verb == "mutate"
                else None
            )
            ftype = self.backend._get_op_ftype(op_args, operator, override_ftype)
            return TypedValue(value, implementation.rtype, ftype)

        def _translate_case(self, expr, switching_on, cases, default, **kwargs):
            def value(df, **kw):
                pd_dtype = self.backend._dtype_to_pd(result_dtype)
                grouper = kw.get("grouper")

                default_ = default.value(df, **kw)

                if not isinstance(default_, pd.Series):
                    result_index = df.index if grouper is None else grouper.result_index
                    result = pd.Series(default_, index=result_index, dtype=pd_dtype)
                else:
                    result = default_

                if switching_on is not None:
                    switching_on_ = switching_on.value(df, **kw)
                else:
                    switching_on_ = None

                for cond, val in cases[::-1]:
                    cond_ = cond.value(df, **kw)
                    val_ = val.value(df, **kw)

                    if switching_on_ is not None:
                        cond_ = switching_on_ == cond_

                    result.mask(cond_, val_, inplace=True)

                return result

            result_dtype, result_ftype = self._translate_case_common(
                expr, switching_on, cases, default, **kwargs
            )
            return TypedValue(value, result_dtype, result_ftype)

        def arranged_window(
            self,
            value: Callable,
            operator: ops.Operator,
            context_kwargs: dict,
            **translate_kwargs,
        ):
            arrange = context_kwargs.get("arrange")
            if arrange is None:
                warnings.warn("Argument 'arrange' is required with SQL backend.")
                return value
            ordering = translate_ordering(self.backend, arrange)

            def arranged_value(df, **kwargs):
                original_index = df.index
                sorted_df = self.backend.sort_df(df, ordering, translate_kwargs)

                # Original grouper can't be used for sorted dataframe
                sorted_grouper = self.backend.df_grouper(sorted_df)
                kwargs["grouper"] = sorted_grouper

                v = value(sorted_df, **kwargs)
                v_with_original_index = v.loc[original_index]
                return v_with_original_index

            return arranged_value

    class AlignedExpressionEvaluator(
        EagerTableImpl.AlignedExpressionEvaluator[TypedValue[pd.Series]]
    ):
        def _translate_col(self, expr, **kwargs):
            df_col_name = expr.table.df_name_mapping[expr.uuid]
            series = expr.table.df[df_col_name]
            dtype = expr.table.cols[expr.uuid].dtype
            return TypedValue(series, dtype)

        def _translate_literal_col(self, expr, **kwargs):
            assert issubclass(expr.backend, PandasTableImpl)
            return expr.typed_value

        def _translate_function(
            self, expr, implementation, op_args, context_kwargs, **kwargs
        ):
            # Drop index to evaluate aligned
            # TODO: Maybe we can somehow restore the index at the end.
            arguments = [
                arg.value.reset_index(drop=True)
                if isinstance(arg.value, pd.Series)
                else arg.value
                for arg in op_args
            ]

            # Check that they have the same length
            series_lengths = {
                len(arg.index) for arg in arguments if isinstance(arg, pd.Series)
            }
            if len(series_lengths) >= 2:
                arg_lengths = [
                    len(arg.index) if isinstance(arg, pd.Series) else 1
                    for arg in arguments
                ]
                raise AlignmentError(
                    f"Arguments for function {expr.name} aren't aligned. Specifically,"
                    f" the inputs are of length {arg_lengths}. Instead they must either"
                    " all be of the same length or of length 1."
                )

            # Compute value
            operator = implementation.operator

            if operator.ftype == OPType.WINDOW and "arrange" in context_kwargs:
                raise NotImplementedError(
                    "The 'arrange' argument for window functions currently is not"
                    " supported for aligned expressions."
                )

            value = implementation(
                *arguments,
                _tbl=None,
                _df=None,
            )
            override_ftype = (
                OPType.WINDOW if operator.ftype == OPType.AGGREGATE else None
            )
            ftype = PandasTableImpl._get_op_ftype(op_args, operator, override_ftype)
            return TypedValue(value, implementation.rtype, ftype)

    class JoinTranslator(Translator[tuple]):
        """
        This translator takes a conjunction (AND) of equality checks and returns
        a tuple of tuple where the inner tuple contains the left and right column
        of the equality checks.
        """

        def _translate(self, expr, **kwargs):
            if isinstance(expr, Column):
                return expr
            if isinstance(expr, FunctionCall):
                if expr.name == "__eq__":
                    c1 = expr.args[0]
                    c2 = expr.args[1]
                    assert isinstance(c1, Column) and isinstance(c2, Column)
                    return ((c1, c2),)
                if expr.name == "__and__":
                    return tuple(itertools.chain(*expr.args))
            raise ExpressionError(
                f"Invalid ON clause element: {expr}. Only a conjunction of equalities"
                " is supported by pandas (ands of equals)."
            )

    #### HELPERS ####

    def sort_df(self, df, ordering: list[OrderingDescriptor], translate_kwargs=None):
        ascending = [o.asc for o in ordering]
        nulls_first = [o.nulls_first for o in ordering]

        if all(nulls_first):
            na_position = "first"
        elif not any(nulls_first):
            na_position = "last"
        else:
            raise NotImplementedError(
                "Pandas sort can't handle different null positions (first / last)"
                " inside a single sort. This can be resolved by splitting the ordering"
                " into multiple arranges."
            )

        # Add temporary columns
        ordering_cols_names = []
        temporary_ordering_cols = {}
        for o in ordering:
            if isinstance(o.order, Column):
                ordering_cols_names.append(self.df_name_mapping[o.order.uuid])
                continue

            # Order is an expression: Create a temporary column
            tmp_name = "tmp_arrange_" + uuid_to_str(uuid.uuid1())
            tmp_col = self.compiler.translate(
                o.order, **(translate_kwargs or {})
            ).value(df)

            ordering_cols_names.append(tmp_name)
            temporary_ordering_cols[tmp_name] = tmp_col

        df = df.copy(deep=False)
        for col_name, col in temporary_ordering_cols.items():
            df[col_name] = col

        df = df.sort_values(
            by=ordering_cols_names,
            ascending=ascending,
            kind="mergesort",
            na_position=na_position,
        )
        df.drop(columns=list(temporary_ordering_cols), inplace=True)

        return df

    def grouped_df(self, df: pd.DataFrame = None):
        if df is None:
            df = self.df

        if self.grouped_by:
            grouping_cols_names = [
                self.df_name_mapping[col.uuid] for col in self.grouped_by
            ]
            return df.groupby(by=list(grouping_cols_names), dropna=False)
        return df

    def df_grouper(self, df: pd.DataFrame = None):
        if self.grouped_by:
            return self.grouped_df(df).grouper
        return None


def fast_pd_convert_dtypes(obj: pd._typing.NDFrameT, **kwargs) -> pd._typing.NDFrameT:
    """
    Faster alternative to the pandas `DataFrame.convert_dtypes` method.
    Unlike the builtin version, this function tries to minimize unnecessary
    copy operations.

    As long as all columns are already of the nullable pandas types, this
    function is a noop.
    """

    if obj.ndim == 1:
        if isinstance(obj.dtype, pd.core.dtypes.dtypes.ExtensionDtype):
            return obj

        inferred_dtype = pandas.core.dtypes.cast.convert_dtypes(obj._values, **kwargs)
        result = obj.astype(inferred_dtype)
        return result

    if not all(
        isinstance(dtype, pd.core.dtypes.dtypes.ExtensionDtype) for dtype in obj.dtypes
    ):
        results = [
            fast_pd_convert_dtypes(col, **kwargs) for col_name, col in obj.items()
        ]

        if len(results) > 0:
            result = pd.concat(results, axis=1, copy=False, keys=obj.columns)
            result = obj._constructor(result)
            result = result.__finalize__(obj, method="convert_dtypes")
            return result

    return obj


#### BACKEND SPECIFIC OPERATORS ################################################


def _pandas_nullable_comparison(lhs, rhs, op):
    """
    Implement comparison operation between series with non-nullable dtypes.
    """
    x = op(lhs, rhs).astype(pd.BooleanDtype(), copy=False)
    if isinstance(lhs, pd.Series):
        x.mask(lhs.isna(), pd.NA, inplace=True)
    if isinstance(rhs, pd.Series):
        x.mask(rhs.isna(), pd.NA, inplace=True)
    return x


with PandasTableImpl.op(ops.Equal()) as op:

    @op("T, const none -> bool")
    def _eq(lhs, rhs):
        assert rhs is None
        return lhs.isna()

    @op("datetime, datetime -> bool")
    def _eq(lhs, rhs):
        return _pandas_nullable_comparison(lhs, rhs, py_operator.eq)


with PandasTableImpl.op(ops.NotEqual()) as op:

    @op("T, const none -> bool")
    def _ne(lhs, rhs):
        assert rhs is None
        return lhs.notna()

    @op("datetime, datetime -> bool")
    def _ne(lhs, rhs):
        return _pandas_nullable_comparison(lhs, rhs, py_operator.ne)


with PandasTableImpl.op(ops.Less()) as op:

    @op("datetime, datetime -> bool")
    def _lt(lhs, rhs):
        return _pandas_nullable_comparison(lhs, rhs, py_operator.lt)


with PandasTableImpl.op(ops.LessEqual()) as op:

    @op("datetime, datetime -> bool")
    def _le(lhs, rhs):
        return _pandas_nullable_comparison(lhs, rhs, py_operator.le)


with PandasTableImpl.op(ops.Greater()) as op:

    @op("datetime, datetime -> bool")
    def _gt(lhs, rhs):
        return _pandas_nullable_comparison(lhs, rhs, py_operator.gt)


with PandasTableImpl.op(ops.GreaterEqual()) as op:

    @op("datetime, datetime -> bool")
    def _ge(lhs, rhs):
        return _pandas_nullable_comparison(lhs, rhs, py_operator.ge)


with PandasTableImpl.op(ops.Pow()) as op:

    @op("int, int -> float")
    def _pow(lhs, rhs):
        if isinstance(lhs, int):
            return lhs ** rhs.astype(pd.Float64Dtype())
        return lhs.astype(pd.Float64Dtype()) ** rhs


with PandasTableImpl.op(ops.RPow()) as op:

    @op("int, int -> float")
    def _rpow(rhs, lhs):
        return _pow(lhs, rhs)


with PandasTableImpl.op(ops.Round()) as op:

    @op.auto
    def _round(x, decimals=0):
        return x.round(decimals=decimals)


with PandasTableImpl.op(ops.IsIn()) as op:

    @op.auto
    def _isin(x, *values):
        y = x.isin(values)
        y.mask(x.isna(), pd.NA, inplace=True)
        return y

    @op("datetime, const datetime... -> bool")
    def _isin(x, *values):
        y = x.isin(values).astype(pd.BooleanDtype(), copy=False)
        y.mask(x.isna(), pd.NA, inplace=True)
        return y


#### String Functions ####


with PandasTableImpl.op(ops.Strip()) as op:

    @op.auto
    def _strip(x):
        return x.str.strip()


with PandasTableImpl.op(ops.StringLength()) as op:

    @op.auto
    def _str_length(x):
        return x.str.len()


with PandasTableImpl.op(ops.Upper()) as op:

    @op.auto
    def _upper(x):
        return x.str.upper()


with PandasTableImpl.op(ops.Lower()) as op:

    @op.auto
    def _upper(x):
        return x.str.lower()


with PandasTableImpl.op(ops.Replace()) as op:

    @op.auto
    def _replace(x, y, z):
        return x.str.replace(y, z)


with PandasTableImpl.op(ops.StartsWith()) as op:

    @op.auto
    def _startswith(x, y):
        return x.str.startswith(y)


with PandasTableImpl.op(ops.EndsWith()) as op:

    @op.auto
    def _endswith(x, y):
        return x.str.endswith(y)


with PandasTableImpl.op(ops.Contains()) as op:

    @op.auto
    def _contains(x, y):
        return x.str.contains(y)


#### Datetime Functions ####


with PandasTableImpl.op(ops.Year()) as op:

    @op.auto
    def _year(x):
        return x.dt.year


with PandasTableImpl.op(ops.Month()) as op:

    @op.auto
    def _month(x):
        return x.dt.month


with PandasTableImpl.op(ops.Day()) as op:

    @op.auto
    def _day(x):
        return x.dt.day


with PandasTableImpl.op(ops.Hour()) as op:

    @op.auto
    def _hour(x):
        return x.dt.hour


with PandasTableImpl.op(ops.Minute()) as op:

    @op.auto
    def _minute(x):
        return x.dt.minute


with PandasTableImpl.op(ops.Second()) as op:

    @op.auto
    def _second(x):
        return x.dt.second


with PandasTableImpl.op(ops.Millisecond()) as op:

    @op.auto
    def _millisecond(x):
        return x.dt.microsecond // 1000


with PandasTableImpl.op(ops.DayOfWeek()) as op:

    @op.auto
    def _day_of_week(x):
        return x.dt.day_of_week + 1


with PandasTableImpl.op(ops.DayOfYear()) as op:

    @op.auto
    def _day_of_year(x):
        return x.dt.day_of_year


#### Generic Functions ####


with PandasTableImpl.op(ops.Greatest()) as op:

    @op.auto
    def _greatest(*x, _df, _result_dtype):
        x_series = []
        x_literals = []

        for e in x:
            if isinstance(e, pd.Series):
                x_series.append(e)
            else:
                x_literals.append(e)

        if x_literals:
            # Calculate max of literals
            literals_max = pd.Series(x_literals, dtype=_result_dtype).max()
            x_series.append(
                pd.Series(literals_max, index=_df.index, dtype=_result_dtype)
            )

        return pd.concat(x_series, copy=False, axis=1).max(axis=1).astype(_result_dtype)


with PandasTableImpl.op(ops.Least()) as op:

    @op.auto
    def _least(*x, _df, _result_dtype):
        x_series = []
        x_literals = []

        for e in x:
            if isinstance(e, pd.Series):
                x_series.append(e)
            else:
                x_literals.append(e)

        if x_literals:
            # Calculate max of literals
            literals_min = pd.Series(x_literals, dtype=_result_dtype).min()
            x_series.append(
                pd.Series(literals_min, index=_df.index, dtype=_result_dtype)
            )

        return pd.concat(x_series, copy=False, axis=1).min(axis=1).astype(_result_dtype)


#### Summarising Functions ####


with PandasTableImpl.op(ops.Mean()) as op:

    @op.auto
    def _mean(x):
        return x.mean()

    @op.auto(variant="transform")
    def _mean(x):
        return x.transform("mean")

    @op.auto(variant="aggregate")
    def _mean(x):
        return x.aggregate("mean")


with PandasTableImpl.op(ops.Min()) as op:

    @op.auto
    def _min(x):
        return x.min()

    @op.auto(variant="transform")
    def _min(x):
        return x.transform("min")

    @op.auto(variant="aggregate")
    def _min(x):
        return x.aggregate("min")


with PandasTableImpl.op(ops.Max()) as op:

    @op.auto
    def _max(x):
        return x.max()

    @op.auto(variant="transform")
    def _max(x):
        return x.transform("max")

    @op.auto(variant="aggregate")
    def _max(x):
        return x.aggregate("max")


with PandasTableImpl.op(ops.Sum()) as op:

    @op.auto
    def _sum(x):
        return x.sum()

    @op.auto(variant="transform")
    def _sum(x):
        return x.transform("sum")

    @op.auto(variant="aggregate")
    def _sum(x):
        return x.aggregate("sum")


with PandasTableImpl.op(ops.Any()) as op:

    @op.auto
    def _any(x):
        return x.any()

    @op.auto(variant="transform")
    def _any(x):
        return x.transform("any")

    @op.auto(variant="aggregate")
    def _any(x):
        return x.aggregate("any")


with PandasTableImpl.op(ops.All()) as op:

    @op.auto
    def _all(x):
        return x.all()

    @op.auto(variant="transform")
    def _all(x):
        return x.transform("all")

    @op.auto(variant="aggregate")
    def _all(x):
        return x.aggregate("all")


with PandasTableImpl.op(ops.StringJoin()) as op:

    @op.auto
    def _join(x, sep):
        # Must remove nulls
        return sep.join(x[x.notnull()])


with PandasTableImpl.op(ops.Count()) as op:

    @op.auto(variant="aggregate")
    @op.auto
    def _count(x: pd.Series):
        # Count non null values
        return x.count()

    @op.auto(variant="transform")
    def _count(x):
        return x.transform("count")


#### Window Functions ####

with PandasTableImpl.op(ops.Shift()) as op:

    @op.auto(variant="transform")
    @op.auto
    def _shift(x: pd.Series, by, empty_value=None):
        return x.shift(by, fill_value=empty_value)


with PandasTableImpl.op(ops.RowNumber()) as op:

    @op.auto
    def _row_number(idx: pd.Series):
        n = len(idx.index)
        return pd.Series(np.arange(1, n + 1), index=idx.index)

    @op.auto(variant="transform")
    def _row_number(idx: pd.Series):
        return idx.cumcount() + 1


with PandasTableImpl.op(ops.Rank()) as op:

    @op.auto
    def _rank(x: pd.Series, *, _ordering: list[OrderingDescriptor]):
        assert len(_ordering) == 1
        ordering = _ordering[0]
        na_option = "top" if ordering.nulls_first else "bottom"

        return x.rank(method="min", ascending=ordering.asc, na_option=na_option)

    # transform variant currently disabled due to bug in pandas:
    # https://github.com/pandas-dev/pandas/issues/54206

    # @op.auto(variant="transform")
    # def _rank(x: pd.Series, *, _ordering: list[OrderingDescriptor]):
    #     assert len(_ordering) == 1
    #     ordering = _ordering[0]
    #     na_option = "top" if ordering.nulls_first else "bottom"
    #     return x.transform(
    #         "rank", method="min", ascending=ordering.asc, na_option=na_option
    #     )


with PandasTableImpl.op(ops.DenseRank()) as op:

    @op.auto
    def _dense_rank(x: pd.Series, *, _ordering: list[OrderingDescriptor]):
        assert len(_ordering) == 1
        ordering = _ordering[0]
        na_option = "top" if ordering.nulls_first else "bottom"

        return x.rank(method="dense", ascending=ordering.asc, na_option=na_option)

    # transform variant currently disabled due to bug in pandas:
    # https://github.com/pandas-dev/pandas/issues/54206

    # @op.auto(variant="transform")
    # def _dense_rank(x: pd.Series, *, _ordering: list[OrderingDescriptor]):
    #     assert len(_ordering) == 1
    #     ordering = _ordering[0]
    #     na_option = "top" if ordering.nulls_first else "bottom"
    #     return x.transform(
    #         "rank", method="dense", ascending=ordering.asc, na_option=na_option
    #     )



from __future__ import annotations

from typing import TYPE_CHECKING, Callable, TypeVar

if TYPE_CHECKING:
    from pydiverse.transform.core.table_impl import AbstractTableImpl


T = TypeVar("T")
ImplT = TypeVar("ImplT", bound="AbstractTableImpl")
CallableT = TypeVar("CallableT", bound=Callable)



from __future__ import annotations

import sqlalchemy as sa

from pydiverse.transform import ops
from pydiverse.transform.lazy.sql_table.sql_table import SQLTableImpl


class PostgresTableImpl(SQLTableImpl):
    _dialect_name = "postgresql"


with PostgresTableImpl.op(ops.Less()) as op:

    @op("str, str -> bool")
    def _lt(x, y):
        return x < y.collate("POSIX")


with PostgresTableImpl.op(ops.LessEqual()) as op:

    @op("str, str -> bool")
    def _le(x, y):
        return x <= y.collate("POSIX")


with PostgresTableImpl.op(ops.Greater()) as op:

    @op("str, str -> bool")
    def _gt(x, y):
        return x > y.collate("POSIX")


with PostgresTableImpl.op(ops.GreaterEqual()) as op:

    @op("str, str -> bool")
    def _ge(x, y):
        return x >= y.collate("POSIX")


with PostgresTableImpl.op(ops.Round()) as op:

    @op.auto
    def _round(x, decimals=0):
        if decimals == 0:
            if isinstance(x.type, sa.Integer):
                return x
            return sa.func.ROUND(x, type_=x.type)

        if isinstance(x.type, sa.Float):
            # Postgres doesn't support rounding of doubles to specific precision
            # -> Must first cast to numeric
            return sa.func.ROUND(sa.cast(x, sa.Numeric), decimals, type_=sa.Numeric)

        return sa.func.ROUND(x, decimals, type_=x.type)


with PostgresTableImpl.op(ops.Second()) as op:

    @op.auto
    def _second(x):
        return sa.func.FLOOR(sa.extract("second", x), type_=sa.Integer())


with PostgresTableImpl.op(ops.Millisecond()) as op:

    @op.auto
    def _millisecond(x):
        _1000 = sa.literal_column("1000")
        return sa.func.FLOOR(sa.extract("milliseconds", x) % _1000, type_=sa.Integer())


with PostgresTableImpl.op(ops.Greatest()) as op:

    @op("str... -> str")
    def _greatest(*x):
        # TODO: Determine return type
        return sa.func.GREATEST(*(e.collate("POSIX") for e in x))


with PostgresTableImpl.op(ops.Least()) as op:

    @op("str... -> str")
    def _least(*x):
        # TODO: Determine return type
        return sa.func.LEAST(*(e.collate("POSIX") for e in x))


with PostgresTableImpl.op(ops.Any()) as op:

    @op.auto
    def _any(x, *, _window_partition_by=None, _window_order_by=None):
        return sa.func.coalesce(sa.func.BOOL_OR(x, type_=sa.Boolean()), sa.false())

    @op.auto(variant="window")
    def _any(x, *, _window_partition_by=None, _window_order_by=None):
        return sa.func.coalesce(
            sa.func.BOOL_OR(x, type_=sa.Boolean()).over(
                partition_by=_window_partition_by,
                order_by=_window_order_by,
            ),
            sa.false(),
        )


with PostgresTableImpl.op(ops.All()) as op:

    @op.auto
    def _all(x):
        return sa.func.coalesce(sa.func.BOOL_AND(x, type_=sa.Boolean()), sa.false())

    @op.auto(variant="window")
    def _all(x, *, _window_partition_by=None, _window_order_by=None):
        return sa.func.coalesce(
            sa.func.BOOL_AND(x, type_=sa.Boolean()).over(
                partition_by=_window_partition_by,
                order_by=_window_order_by,
            ),
            sa.false(),
        )



from __future__ import annotations

from .aggregate import *
from .core import *
from .datetime import *
from .generic import *
from .logical import *
from .markers import *
from .numeric import *
from .string import *
from .window import *



from __future__ import annotations

from .expressions import (
    CaseExpression,
    Column,
    FunctionCall,
    LambdaColumn,
    LiteralColumn,
    expr_repr,
)
from .symbolic_expressions import SymbolicExpression, unwrap_symbolic_expressions
from .translator import Translator, TypedValue
from .util import iterate_over_expr
